//******************************************************************************
// Written by   : Cagatay Undeger
// Date         : 26.02.2004
// Version      : 2.0
//******************************************************************************

#ifndef __GAME_ENVIRONMENT_MESH_LOADER__
#define __GAME_ENVIRONMENT_MESH_LOADER__

#include "vector_operations.h"
#include "ImageLoader.h"

//********************************************************
// Classes
//********************************************************

class TTriangle;
class TMeshBoundingBox;
class TMeshFrame;
class TMeshPart;
class TMeshMaterial;
class TMesh;
class TMeshLoader;

//********************************************************
// Externals
//********************************************************

 extern TMeshLoader MeshLoader;

//********************************************************
// TMeshMaterial
//********************************************************

class  TMeshMaterial
{
        friend TMesh;
        friend TMeshPart;
        friend TMeshLoader;
protected:
        TMesh         * mesh;
        mstring         name;
        int             materialtype;

        mstring         diffusefile;
        mstring         lastdiffusefileloaded;
        bool		lastclonetex;
        TTexture      * diffusetexture;
public:
        TMeshMaterial  ( );
        ~TMeshMaterial ( );

        TMesh         * GetMesh           ( );
        mstring         GetName           ( );
        int             GetMaterialType   ( );
        mstring         GetDiffuseFileName( );
        TTexture      * GetDiffuseTexture ( );

        void            SetName           ( mstring pname );
        void            SetMaterialType   ( int mtype );
        void            SetDiffuseFileName( mstring fname );

        void            LoadMaterial    ( );
};

//********************************************************
// TMeshFrame
//********************************************************

class  TMeshFrame
{
        friend TMesh;
        friend TMeshPart;
        friend TMeshLoader;
protected:
        TMesh                 * mesh;
        mstring                 name;
	std::vector<TVertex3f>	vertices;
	std::vector<TVertex3f>	normals;
	std::vector<TColor3f>	diffuses;
	std::vector<TColor3f>	speculars;
	std::vector<float>	shininess;
	std::vector<float>	alphas;
	std::vector<TVertex2f>	mappingcoords;
public:
        TMeshFrame  ( );
        ~TMeshFrame ( );

        TMesh          * GetMesh ( );
        mstring          GetName ( );

        void             SetName ( mstring pname );

        int              GetNumOfVertices  ( );
        TVertex3f *      GetVertex         ( unsigned int index );
        TVertex3f *      AddVertex         ( unsigned int * index = NULL );
        void             DeleteVertex      ( unsigned int index );

        int              GetNumOfNormals   ( );
        TVertex3f *      GetNormal         ( unsigned int index );
        TVertex3f *      AddNormal         ( unsigned int * index = NULL );
        void             DeleteNormal      ( unsigned int index );


        int              GetNumOfDiffuses  ( );
        TColor3f *       GetDiffuse        ( unsigned int index );
        TColor3f *       AddDiffuse        ( unsigned int * index = NULL );
        void             DeleteDiffuse     ( unsigned int index );

        int              GetNumOfSpeculars  ( );
        TColor3f *       GetSpecular        ( unsigned int index );
        TColor3f *       AddSpecular        ( unsigned int * index = NULL );
        void             DeleteSpecular     ( unsigned int index );

        int              GetNumOfShininess ( );
        float *          GetShininess      ( unsigned int index );
        float *          AddShininess      ( unsigned int * index = NULL );
        void             DeleteShininess   ( unsigned int index );

        int              GetNumOfAlphas    ( );
        float *          GetAlpha          ( unsigned int index );
        float *          AddAlpha          ( unsigned int * index = NULL );
        void             DeleteAlpha       ( unsigned int index );

        int              GetNumOfMappingCoords ( );
        TVertex2f *      GetMappingCoord       ( unsigned int index );
        TVertex2f *      AddMappingCoord       ( unsigned int * index = NULL );
        void             DeleteMappingCoord    ( unsigned int index );
};

//********************************************************
// TMeshBoundingBox
//********************************************************

class  TMeshBoundingBox
{
public:
        float centerx, centery, centerz;
        float sizex, sizey, sizez, sizexydiagonal, sizediagonal;
        float minx,miny,minz;
        float maxx,maxy,maxz;
        TMeshBoundingBox ();
};

//********************************************************
// TTriangle
//********************************************************

class  TTriangle
{
public:
        int     vertices        [3];
        int     normals         [3];
        int     mappingcoords   [3];
	int     diffuse;
	int     specular;
	int     shininess;
	int     alpha;
        int     materialID;
        TTriangle ();
};

//********************************************************
// TMeshPart
//********************************************************

class  TMeshPart
{
        friend TMesh;
        friend TMeshLoader;
protected:
        TMesh          * mesh;
        mstring          name;
        float            nearplane;
        float            farplane;
        bool             hbillboard;
        bool             pbillboard;
        bool             rbillboard;
        TColor3f         emission;
        float            visiblealphathreshold;
        bool             doublefaced;
        bool             visible;
        float            alpha;

        TMeshBoundingBox bbox;

	std::vector<TTriangle> triangles;
public:
        TMeshPart  ( );
        ~TMeshPart ( );

        TMesh          * GetMesh ( );
        mstring          GetName ( );
        float            GetNearplane   ( );
        float            GetFarPlane    ( );
        bool             GetHBillboard  ( );
        bool             GetPBillboard  ( );
        bool             GetRBillboard  ( );
        TColor3f         GetEmission    ( );
        bool             GetDoubleFaced ( );
        float            GetAlpha       ( );
        bool             GetVisible     ( );
        float            GetVisibleAlphaThreshold ( );

        TMeshBoundingBox * GetBBox  ( );
        void             UpdateBBox ( );

        void             SetName        ( mstring pname );
        void             SetNearPlane   ( float np );
        void             SetFarPlane    ( float fp );
        void             SetHBillboard  ( bool billb );
        void             SetPBillboard  ( bool billb );
        void             SetRBillboard  ( bool billb );
        void             SetEmission    ( TColor3f emis  );
        void             SetDoubleFaced ( bool douface );
        void             SetAlpha       ( float alph );
        void             SetVisible     ( bool vis );
        void             SetVisibleAlphaThreshold ( float vist );

        int              GetNumOfTriangles ( );
        TTriangle *      GetTriangle       ( unsigned int index );
        TTriangle *      AddTriangle       ( unsigned int * index = NULL );
        void             DeleteTriangle    ( unsigned int index );

        bool             IsTransparent     ( );

        void             Render            ( float camh, float camp, float camr,
                                             float objh, float objp, float objr,
                                             float camrange, float globalalpha,
                                             bool  hasanytransparent,
                                             int &prevtexture_enabled,
                                             TTexture * &prevtexture,
                                             bool notexture = false,
                                             bool usefixedcolor = false,
                                             TColor3f fixedcolor = TColor3f(1,1,1),
                                             bool disablebackfaceculling = false );
};

//********************************************************
// TMesh
//********************************************************

class  TMesh
{
        friend TMeshLoader;
        friend TMeshPart;
        friend TMeshFrame;
        friend TMeshMaterial;
protected:
        mstring          filename;
        unsigned int     numoflinks;
        bool             clone;
        bool             clonetexture;

        unsigned int     currentframe;
        unsigned int     targetframe;
        float            frametransition;
        bool             smoothframetransitions;

        float            alpha;
        void           * data;
        int              istransparent;

        TMeshBoundingBox bbox;

        std::vector<TMeshFrame*>        frames;
	std::vector<TMeshPart*>         parts;
	std::vector<TMeshMaterial*>     materials;
public:
        TMesh  ( );
        ~TMesh ( );

        mstring          GetFullFilePath   ( );
        mstring          GetFullFileName   ( );
        mstring          GetOnlyFileName   ( );
        void             SetFileName       ( mstring fname );

        unsigned int     GetNumOfFrames    ( );
        TMeshFrame     * GetFrame          ( unsigned int index );
        TMeshFrame     * GetCurrentFrame   ( );
        TMeshFrame     * GetTargetFrame    ( );
        unsigned int     GetCurrentFrameNo ( );
        unsigned int     GetTargetFrameNo  ( );
        float            GetFrameTransition( );
        bool             GetSmoothFrameTransitions ( );

        TMeshFrame     * AddFrame          ( unsigned int * index = NULL );
        void             DeleteFrame       ( unsigned int index );

        void             SetCurrentFrameNo ( unsigned int index );
        void             SetTargetFrameNo  ( unsigned int index );
        void             SetFrameTransition( float trns );
        void             SetSmoothFrameTransitions ( bool smt );

        void             LoadMaterials          ( );
        void             AdvanceMaterialAnims   ( );

        TMeshBoundingBox * GetBBox              ( );
        void             UpdateBBox             ( );
        void             UpdateBBoxFromSubParts ( );

        float            GetAlpha          ( );
        void           * GetData           ( );

        void             SetAlpha          ( float palpha );
        void             SetData           ( void * pdata );

        unsigned int     GetNumOfParts     ( );
        TMeshPart      * GetPart           ( unsigned int index );
        TMeshPart      * AddPart           ( unsigned int * index = NULL );
        void             DeletePart        ( unsigned int index );

        unsigned int     GetNumOfMaterials ( );
        TMeshMaterial  * GetMaterial       ( unsigned int index );
        TMeshMaterial  * AddMaterial       ( unsigned int * index = NULL );
        void             DeleteMaterial    ( unsigned int index );

        int              GetNumOfTriangles ( );

        bool             IsTransparent     ( );

        void             FitInto           ( bool proportional, float minx, float miny, float minz,
                                             float maxx, float maxy, float maxz );

        void             ClearAll          ( );

        bool             SaveToStream ( TGameIOStream * stream );

        bool             LoadFromFile      ( mstring fname,
                                             bool append_frame_only = false,
                                             unsigned int frame_no = 0,
                                             float nearplane = 0,
                                             float farplane  = 3000,
                                             void * LoaderParameters = NULL );

        bool             LoadFromStream ( TGameIOStream * stream,
                                             bool append_frame_only = false,
                                             unsigned int frame_no = 0,
                                             float nearplane = 0,
                                             float farplane  = 3000,
                                             void * LoaderParameters = NULL );

        bool             SaveToFile        ( mstring fname );

        void             Render            ( float camh, float camp, float camr,
                                             float objh, float objp, float objr,
                                             float camrange,
                                             float globalalpha,
                                             int  &prevtexture_enabled,
                                             TTexture * &prevtexture,
                                             bool notexture = false,
                                             bool usefixedcolor = false,
                                             TColor3f fixedcolor = TColor3f(1,1,1),
                                             bool disablebackfaceculling = false 
                                           );
};

//********************************************************
// TMeshLoader
//********************************************************

class  TMeshLoader
{
protected:
        std::vector<TMesh*> meshes;
        void DeleteMeshByIndex ( int index );
public:

        TMeshLoader();
        virtual ~TMeshLoader ();

        TMesh *    LinkMesh             ( mstring fname,
                                          bool clone = false, bool clone_texture = false,
                                          float nearplane = 0, float farplane  = 3000,
                                          void * LoaderParameters = NULL,
                                          mstring rootpath = ApplicationDirectory );
        void       UnlinkMesh           ( TMesh * mesh);

        void       ClearAll             ( );
        void       DeleteMesh           ( TMesh * mesh );
        void       DetachMesh           ( TMesh * mesh );
};

#endif
