#include "environment_visualizer.h"

#include <math.h>

#include <float.h>

#include <GL\gl.h>
#include <GL\glu.h>

#include "graphics_utilities.h"
#include "extgl.h"

//******************************************************************************
// TEnvironmentVisualizer
//******************************************************************************

TEnvironmentVisualizer::TEnvironmentVisualizer ()
{
        fillpoly        = true;
        smoothshade     = true;
        gllighting      = true;

        lasttime        = 0;
        frametime       = 0.1f;

        widthpixels     = 320;
        heightpixels    = 240;

        sunheading      = -60;
        sunpitch        =  50;

        clearColorR     = 69.0/255;
        clearColorG     = 149.0/255;
        clearColorB     = 231.0/255;
}

TEnvironmentVisualizer::~TEnvironmentVisualizer ()
{
}

bool TEnvironmentVisualizer::GetFilledTriangles       ( )
{
        return fillpoly;
}

bool TEnvironmentVisualizer::GetLightingOn            ( )
{
        return gllighting;
}

void TEnvironmentVisualizer::SetFilledTriangles         ( bool val )
{
        fillpoly = val;
}

void TEnvironmentVisualizer::SetLightingOn              ( bool val )
{
        gllighting = val;
}

float TEnvironmentVisualizer::GetFrameTime ( )
{
        return frametime;
}

int TEnvironmentVisualizer::GetScreenWidth	           ( )
{
        return widthpixels;
}

int TEnvironmentVisualizer::GetScreenHeight                 ( )
{
        return heightpixels;
}

void TEnvironmentVisualizer::SetScreenWidth	           ( int sw   )
{
        widthpixels = sw;
}

void TEnvironmentVisualizer::SetScreenHeight                 ( int sh   )
{
        heightpixels = sh;
}

float TEnvironmentVisualizer::GetSunHeading        ( )
{
        return sunheading;
}

void  TEnvironmentVisualizer::SetSunHeading        ( float sh )
{
        sunheading = sh;
}

float TEnvironmentVisualizer::GetSunPitch          ( )
{
        return sunpitch;
}

void  TEnvironmentVisualizer::SetSunPitch          ( float sp )
{
        sunpitch = sp;
}

void    TEnvironmentVisualizer::GetBackgroundColor   ( float &r, float &g, float &b )
{
        r = clearColorR;
        g = clearColorG;
        b = clearColorB;
}

void    TEnvironmentVisualizer::SetBackgroundColor   ( float r,  float g,  float b  )
{
        clearColorR     = r;
        clearColorG     = g;
        clearColorB     = b;
}

void TEnvironmentVisualizer::InitializeRenderingPers ( float x, float y, float z, float h, float p, float r,
                      			                        float vfov, float nearplane, float farplane,
                                                                bool clearbackground,
                                                                bool cleardepthbuffer )
{
        //------------- Configure Screen -------------------

        #ifdef __linux__
                // NOT COMPLETED
        #else
                _controlfp(MCW_EM, MCW_EM);
                _controlfp(CW_DEFAULT ,0xfffff);
        #endif

        glViewport(0, 0, widthpixels, heightpixels );

        if (clearbackground && cleardepthbuffer)
        {
                glClearColor( clearColorR, clearColorG, clearColorB, 1 );
                glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
        }
        else
        if (cleardepthbuffer)
        {
                glClear(GL_DEPTH_BUFFER_BIT);
        }
        else
        {
                glClearColor( clearColorR, clearColorG, clearColorB, 1 );
                glClear(GL_COLOR_BUFFER_BIT);
        }

        float screen_width_div_height = float(widthpixels)/heightpixels;
        float hfov  = atan( screen_width_div_height*tan(vfov/360.0*PI) )*360.0/PI;

        camx 	= x;
        camy 	= y;
        camz 	= z;
        camh 	= h;
        camp 	= p;
        camr 	= r;
        camhfov = hfov;
        camvfov = vfov;
        camfarplane = farplane;
        clearback   = clearbackground;

        //------------- Compute Frame Time -------------------

        unsigned int ctime = mGetTickCount();
        if (lasttime)
        {
                frametime = 0.5*frametime + 0.5*(ctime-lasttime)/1000.0f;
                if ( frametime < 0.001f ) frametime = 0.001f;
        }
        lasttime = ctime;

        //------------- Store Initial GL State ---------------

        glPushAttrib ( GL_ALL_ATTRIB_BITS );

        //------------- Set GL State -------------------------

        if (gllighting)
        	glEnable ( GL_LIGHTING );
        glEnable         ( GL_DEPTH_TEST );
        glDepthFunc      ( GL_LEQUAL );
        glEnable         ( GL_TEXTURE_2D );
        glEnable         ( GL_COLOR_MATERIAL );
        glEnable         ( GL_CULL_FACE );
        glCullFace       ( GL_BACK  );
        glEnable         ( GL_BLEND );
        glBlendFunc      ( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

        float ambient  [] = { 0,0,0,0 };

        glLightModelfv( GL_LIGHT_MODEL_AMBIENT, ambient );

        float color    [] = { 1,1,1,1 };
        float emission [] = { 0,0,0,0 };
        float specular [] = { 0,0,0,0 };
        float shininess   = 0;

        glMaterialfv( GL_FRONT_AND_BACK,GL_AMBIENT,  color );
        glMaterialfv( GL_FRONT_AND_BACK,GL_DIFFUSE,  color );
        glMaterialfv( GL_FRONT_AND_BACK,GL_EMISSION, emission );
        glMaterialfv( GL_FRONT_AND_BACK,GL_SPECULAR, specular );
        glMaterialf ( GL_FRONT_AND_BACK,GL_SHININESS,shininess );

        //------------- Initialize Camera --------------------

        glMatrixMode   ( GL_PROJECTION );
        glLoadIdentity ( );
        gluPerspective ( vfov, screen_width_div_height, nearplane, camfarplane );

	glMatrixMode   ( GL_MODELVIEW );
        glLoadIdentity ( );

        glRotatef      ( -90, 1, 0, 0 );
        glRotatef      ( -r,  0, 1, 0 );
        glRotatef      ( -p,  1, 0, 0 );
        glRotatef      (  h,  0, 0, 1 );

        glTranslatef   (  -x,  -y, -z   );

        //------------- Initialize Render Mode  --------------

        if (fillpoly)
                glPolygonMode ( GL_FRONT_AND_BACK, GL_FILL );
                else
                glPolygonMode ( GL_FRONT_AND_BACK, GL_LINE );

        if (smoothshade)
                glShadeModel ( GL_SMOOTH );
                else
                glShadeModel ( GL_FLAT );
}

void TEnvironmentVisualizer::InitializeLights ( )
{
        if ( !gllighting ) return;

        double dx, dy, dz;
        double sourcex, sourcey, sourcez, sourceh, sourcep;

        //**********************************************************************
        // compute location of sun
        //**********************************************************************

        dx = 0;
        dy = camfarplane*0.9;
        dz = 0;

        sourceh = GetSunHeading( );
        sourcep = GetSunPitch( );

        VectorOp.Rotate_X (  sourcep, &dy, &dz );
        VectorOp.Rotate_Z ( -sourceh, &dx, &dy );

        //**********************************************************************
        // set sun light as first light source
        //**********************************************************************

        // openGL lighting variables
        unsigned int lno;
	float quad_fac;

	float glfLightPosition    [] = {   0,    0,    0,   1 };
	float glfLightAmbient     [] = {   0,    0,    0,   1 };
   	float glfLightDiffuse     [] = {   0,    0,    0,   1 };
   	float glfLightSpecular    [] = {   0,    0,    0,   1 };

        // initialize lighting for sun
        lno   = GL_LIGHT0;

        glEnable ( lno );
        glLightf ( lno, GL_CONSTANT_ATTENUATION , 1    );
        glLightf ( lno, GL_LINEAR_ATTENUATION   , 0    );
        glLightf ( lno, GL_SPOT_CUTOFF, 180 );

        // set color of sun/moon light
        glfLightDiffuse [0] = 1.0;
        glfLightDiffuse [1] = 1.0;
        glfLightDiffuse [2] = 1.0;

        glfLightSpecular[0] = glfLightDiffuse[0];
        glfLightSpecular[1] = glfLightDiffuse[1];
        glfLightSpecular[2] = glfLightDiffuse[2];

        glfLightAmbient [0] = 0.1;
        glfLightAmbient [1] = 0.1;
        glfLightAmbient [2] = 0.1;

        glLightfv( lno, GL_AMBIENT , glfLightAmbient );
        glLightfv( lno, GL_DIFFUSE , glfLightDiffuse );
        glLightfv( lno, GL_SPECULAR, glfLightSpecular);

        // set position of sun/moon light
        dx =   0;
        dy =   1;
        dz =   0;
        VectorOp.Rotate_X (  sourcep, &dy, &dz );
        VectorOp.Rotate_Z ( -sourceh, &dx, &dy );

        glfLightPosition[0] = dx;
        glfLightPosition[1] = dy;
        glfLightPosition[2] = dz;
        glfLightPosition[3] = 0;
        glLightfv(lno, GL_POSITION, glfLightPosition);
}

void TEnvironmentVisualizer::RenderObjects 	   ( TGameObjectLayer * layer )
{
        for ( int i = 0; i<layer->GetNumOfObjects(); i++ )
        {
                TGameObject     * std  = layer->GetObject(i);
                if ( std->GetVisible() && std->IsAClass( "TGameMeshObject" ) )
                {
                        TGameMeshObject * obj  = (TGameMeshObject *)std;
                        TMesh *           mesh = obj->GetMesh();
                        if ( mesh )
                        {
                                int prevtexture_enabled = 0;
                                TTexture * prevtexture = NULL;
                                mesh->Render( camh, camp, camr,
                                              obj->GetH(), obj->GetP(), obj->GetR(),
                                              0, obj->GetOpacity(),
                                              prevtexture_enabled, prevtexture,
                                              false, false, TColor3f(1,1,1),
                                              !obj->GetBackFaceCulling() );
                        }
                        obj->CustomView3dGL( camh, camp, camr, camhfov, camvfov, camfarplane,
                                             widthpixels, heightpixels );
                }
        }
}

void TEnvironmentVisualizer::FinalizeRendering ( )
{
        glPopAttrib   ( );
}


