#include "game_plugins.h"

//******************************************************************************
// TGameFileMeshObject
//******************************************************************************

TGameFileMeshObject::TGameFileMeshObject ( TGameObjectLayer * player ) : TGameMeshObject ( player )
{
        AddClassName ( "TGameFileMeshObject" );

        meshFile                = "";
        animationOn             = true;
        useAllAnimationFrames   = true;
        animationStartFrame     = 0;
        animationEndFrame       = 0;
        animationFrameDuration  = 40;
        currentFrameTime        = 0;
        loopAnimation           = true;
}

TGameFileMeshObject::~TGameFileMeshObject ( )
{
        DeallocateMesh ( );
}

void     TGameFileMeshObject::SetUniqueMesh                   ( bool unique )
{
        if ( uniqueMesh == unique ) return;
        DeallocateMesh( );
        TGameMeshObject::SetUniqueMesh ( unique );
        GenerateMesh  ( );
}

void TGameFileMeshObject::AdvanceTime        ( unsigned int deltatime )
{
        TGameMeshObject::AdvanceTime ( deltatime );
        if ( !animationOn || !mesh ) return;

        // advance simulation time
        currentFrameTime += deltatime;

        // find frame bounds
        int startframe, endframe;
        if ( useAllAnimationFrames )
        {
                startframe = 0;
                endframe   = mesh->GetNumOfFrames();
        }
        else
        {
                startframe = animationStartFrame;
                endframe   = animationEndFrame;
        }

        // find current frame
        while ( currentFrameTime >= animationFrameDuration )
        {
                currentFrameTime -= animationFrameDuration;
                if ( currentFrame<startframe || currentFrame>endframe )
                        currentFrame = startframe;
                        else
                        currentFrame++;
                if ( loopAnimation )
                {
                        if ( currentFrame > endframe )
                                currentFrame = startframe;
                }
                else
                {
                        if ( currentFrame > endframe )
                                currentFrame = endframe;
                }
        }

        // find target frame
        if ( currentFrame < startframe )
                targetFrame = startframe;
                else
                targetFrame = currentFrame+1;
        if ( loopAnimation )
        {
                if ( targetFrame > endframe )
                        targetFrame = startframe;
        }
        else
        {
                if ( targetFrame > endframe )
                        targetFrame = endframe;
        }

        // find frameTransitionRatio
        frameTransitionRatio = float(currentFrameTime)/animationFrameDuration;
}

// save and load
bool 		TGameFileMeshObject::LoadFromStream 	   ( TGameIOStream * stream )
{
	TGameMeshObject::LoadFromStream ( stream );
        int version = 1;
        stream->Read ( &version );

        stream->ReadString ( &meshFile );
        stream->Read ( &animationOn );
        stream->Read ( &useAllAnimationFrames );
        stream->Read ( &animationStartFrame );
        stream->Read ( &animationEndFrame );
        stream->Read ( &animationFrameDuration );
        stream->Read ( &currentFrameTime );
        stream->Read ( &loopAnimation );

        if (uniqueMesh)
        {
                if ( mesh )
                        mesh->SetFileName ( meshFile );
        }
        else
        GenerateMesh ( );

        return true;
}

bool 		TGameFileMeshObject::SaveToStream   	   ( TGameIOStream * stream )
{
	TGameMeshObject::SaveToStream ( stream );
        int version = 1;
        stream->Write ( &version );

        stream->WriteString ( &meshFile );
        stream->Write ( &animationOn );
        stream->Write ( &useAllAnimationFrames );
        stream->Write ( &animationStartFrame );
        stream->Write ( &animationEndFrame );
        stream->Write ( &animationFrameDuration );
        stream->Write ( &currentFrameTime );
        stream->Write ( &loopAnimation );

        return true;
}

//*************** additional methods ****************

void TGameFileMeshObject::DeallocateMesh ( )
{
        TGameMeshObject::DeallocateMesh ( );
        currentFrame          = 0;
        targetFrame           = 0;
        frameTransitionRatio  = 0;
	animationStartFrame   = 0;
        animationEndFrame     = 0;
}

void TGameFileMeshObject::GenerateMesh  ( )
{
        DeallocateMesh();
        mesh = MeshLoader.LinkMesh( meshFile, uniqueMesh, false );
        if ( uniqueMesh )
                MeshLoader.DetachMesh ( mesh );
        animationEndFrame     = mesh->GetNumOfFrames()-1;
        GenerateBoundingVolumes ( );
}

void    TGameFileMeshObject::GenerateBoundingVolumes ( )
{
        boundingVolumes.clear();
        if ( mesh )
        {
                TMeshBoundingBox * meshbb;
                TGameMeshBoundingVolume bv;
                for ( int i = 0; i<mesh->GetNumOfParts(); i++ )
                {
                        meshbb = mesh->GetPart(i)->GetBBox();

                        bv.type         = 0; // box

                        bv.centerx      = meshbb->centerx;
                        bv.centery      = meshbb->centery;
                        bv.centerz      = meshbb->centerz;
                        bv.sizex        = meshbb->sizex;
                        bv.sizey        = meshbb->sizey;
                        bv.sizez        = meshbb->sizez;
                        bv.minx         = meshbb->minx;
                        bv.miny         = meshbb->miny;
                        bv.minz         = meshbb->minz;
                        bv.maxx         = meshbb->maxx;
                        bv.maxy         = meshbb->maxy;
                        bv.maxz         = meshbb->maxz;

                        boundingVolumes.push_back(bv);
                }
        }
        GenerateEntireBoundingVolume ();
}

mstring         TGameFileMeshObject::GetMeshFile                     ( )
{
        return meshFile;
}

void            TGameFileMeshObject::SetMeshFile                     ( mstring mfile )
{
        if ( meshFile == mfile ) return;
        meshFile = mfile;
        GenerateMesh ( );
}

bool            TGameFileMeshObject::GetAnimationOn                  ( )
{
        return animationOn;
}

void            TGameFileMeshObject::SetAnimationOn                  ( bool animon )
{
        animationOn = animon;
}

bool		TGameFileMeshObject::GetUseAllAnimationFrames        ( )
{
        return useAllAnimationFrames;
}

void            TGameFileMeshObject::SetUseAllAnimationFrames        ( bool useallf )
{
        useAllAnimationFrames = useallf;
}

unsigned int	TGameFileMeshObject::GetAnimationStartFrame          ( )
{
        return animationStartFrame;
}

void            TGameFileMeshObject::SetAnimationStartFrame          ( unsigned int frm )
{
        animationStartFrame = frm;
}

unsigned int	TGameFileMeshObject::GetAnimationEndFrame            ( )
{
        return animationEndFrame;
}

void            TGameFileMeshObject::SetAnimationEndFrame            ( unsigned int frm )
{
        animationEndFrame = frm;
}

unsigned int	TGameFileMeshObject::GetAnimationFrameDuration       ( )
{
        return animationFrameDuration;
}

void            TGameFileMeshObject::SetAnimationFrameDuration       ( unsigned int fduration )
{
        if ( fduration == 0 ) fduration = 1;
        animationFrameDuration = fduration;
}

bool		TGameFileMeshObject::GetLoopAnimation                ( )
{
        return loopAnimation;
}

void            TGameFileMeshObject::SetLoopAnimation                ( bool loopa )
{
        loopAnimation = loopa;
}

//******************************************************************************
// TGameTerrainObject
//******************************************************************************

TGameTerrainObject::TGameTerrainObject ( TGameObjectLayer * player ) : TGameFileMeshObject( player )
{
        AddClassName ( "TGameTerrainObject" );
}

TGameTerrainObject::~TGameTerrainObject ( )
{
}

//******************************************************************************
// TGameRiverObject
//******************************************************************************

TGameRiverObject::TGameRiverObject ( TGameObjectLayer * player ) : TGameFileMeshObject( player )
{
        AddClassName ( "TGameRiverObject" );
}

TGameRiverObject::~TGameRiverObject ( )
{
}

